# -*- coding: utf-8 -*-
"""
Created on Mon Jan 25 14:29:51 2021

@author: j.reul

The module incorporates the class "Core", which inherits functionality from the 
classes "Estimation", "Simulation", "PostAnalysis" and "Visualitation".
It defines the core attributes, structure and type of the
discrete choice models to be simulated or estimated. 
Two different types of discrete choice models are differentiated:
    - Multinomial logit models (MNL)
    - Mixed logit models (MXL) with discrete points as parameters

"""

import numpy as np
import pandas as pd
import pickle
import os

from .estimation import Estimation
from .post_analysis import PostAnalysis
from .simulation import Simulation
from . import config

class Core(Estimation, Simulation, PostAnalysis):
    """
    The class "Core" defines the attributes, structure and type of the 
    mixed logit model, being built.
    
    The class inherits functionality from the classes "Estimation"
    and "Simulation"
    """
    
    def __init__(self, **kwargs):
        """
        Object initialization of class "Core".
             
        Parameters
        ----------
        kwargs param : dict
            Holds the names of all attributes in utility function.
            list(param) = ['constant', 'variable'] --> Disctinction between variables
            that are constant over alternatives or vary, respectively.
            list(param['constant']) = ['fixed', 'random'] --> Distinction
            between variables that are not randomly distributed within
            a group of decision-makers ('fixed') and those, that are randomly
            distributed with a discrete distribution ('random')
                                    
        kwargs max_space : int
            Maximum number of data points within parameter space.
            
        kwargs alt: int
            Number of discrete choice alternatives.
            
        kwargs equal_alt: int
            Maximum number of equal alternatives (with different attributes) 
            in a single observation/choice set. 
            E.g. mode choice: Available are two buses and one car. This would
            lead to a maximum of two equal alternatives (two buses).
                        
        kwargs norm_alt : int
            Defines, which alternative shall be normalized. Defaults to 0.
                        
        kwargs include_weights : boolean
            If True, the model searches for a column in the input data, which
            is called "weight". This column indicates the weight of each 
            observation in the input data. Defaults to True.
                        
        kwargs data_name : str
            Defines the filename of the file, which holds the base data
            to calibrate the MNL or MXL model on.
        
        kwargs data_index : array
            An array, which holds the index values of those datapoints
            of the base data, which shall be used for estimation.
        
        initial_point_name : str
            The filename of the file, which holds the estimated MNL parameters.
                    
        kwargs dc_type : str
            Determines the model type for simulation: MNL or MXL model. 
            Depends on which estimated model parameters are available.
                
        kwargs dict_specific_travel_cost : dictionary
            External specification of transport costs. Relevant for the 
            simulation of mode choice.
        
        Returns
        -------
        None.

        """
        
        self.model_type = kwargs.get('model_type', 'estimation')
        
        if self.model_type == 'simulation':
                
            self.asc_offset_hh_cars = config.asc_offset_hh_cars
            
            #load previously estimated model parameters, 
            #if model-type is simulation.
            self.initial_point_cars = config.initial_point_cars
            self.initial_point_mode = config.initial_point_mode
                
            self.log_param = config.log_param
            dict_specific_travel_cost_ext = kwargs.get('dict_specific_travel_cost', {})
            cc_cost_ext = kwargs.get('cc_cost', False)
            if len(dict_specific_travel_cost_ext) > 0:
                self.dict_specific_travel_cost = dict_specific_travel_cost_ext
            else:
                self.dict_specific_travel_cost = config.dict_specific_travel_cost
            if cc_cost_ext:
                self.cc_cost = cc_cost_ext
            else:
                self.cc_cost = config.cc_cost_2020
            
            self.param = kwargs.get('param', {})
            if self.param:
                self.no_constant_fixed = len(self.param['constant']['fixed'])
                self.no_constant_random = len(self.param['constant']['random'])
                self.no_variable_fixed = len(self.param['variable']['fixed'])
                self.no_variable_random = len(self.param['variable']['random'])
                
        else:
            #define path to input data
            PATH_MODULE = os.path.dirname(__file__)
            sep = os.path.sep
            self.data_name = kwargs.get("data_name", False)
            self.initial_point_name = kwargs.get("initial_point_name", False)
            self.PATH_InputData = PATH_MODULE + sep + 'InputData' + sep
            self.PATH_ModelParam = PATH_MODULE + sep + 'ModelParam' + sep
            self.PATH_Visualize = PATH_MODULE + sep + 'Visualizations' + sep
            
            #random or fixed within parameter space of Mixed Logit
            self.param = kwargs.get('param', False)
            self.count_c = kwargs.get('alt', False)
            self.count_e = kwargs.get('equal_alt', False)
            
            #check, if necessary values have been specified.
            if self.param == False:
                raise ValueError('Argument -param- needs to be specified!')                
            if self.count_c == False:
                raise ValueError('Argument -alt- needs to be specified!')
            if self.count_e == False:
                raise ValueError('Argument -equal_alt- needs to be specified!')             
            if self.data_name == False:
                raise ValueError('Argument -data_name- needs to be specified!')                
            
            self.no_constant_fixed = len(self.param['constant']['fixed'])
            self.no_constant_random = len(self.param['constant']['random'])
            self.no_variable_fixed = len(self.param['variable']['fixed'])
            self.no_variable_random = len(self.param['variable']['random'])
                            
            self.data_index = kwargs.get("data_index", np.array([]))

            self.include_weights = kwargs.get("include_weights", True)
            
            if self.initial_point_name:
                with open(self.PATH_ModelParam + self.initial_point_name + ".pickle", 'rb') as handle:
                    self.initial_point = pickle.load(handle)  
                                
            print('Data wrangling.')
            
            try:
                self.data = pd.read_csv(self.PATH_InputData + self.data_name + ".csv", sep = ",")
                if len(list(self.data)) == 1:
                    raise ValueError('Check the separator of the imported .csv-files. Should be ","-separated.')
            except:
                with open(self.PATH_InputData + self.data_name + ".pickle", 'rb') as handle:
                    self.data = pickle.load(handle)  
                    
            self.data = self.data.reset_index(drop=True)
                        
            #get data-points from indicated indices
            if self.data_index.size:
                self.data = self.data.iloc[self.data_index]
                                
            print('Length of dataset:', len(self.data))
                                            
            #define choices and availabilities
            #scale availabilities, if weights are provided in input-data
            if "weight" in self.data.columns and self.include_weights == True:
                self.weight_vector = self.data["weight"].values.copy()
            else:
                self.weight_vector = np.ones(shape=len(self.data), dtype=np.float64)
            self.choice = np.zeros((self.count_c,self.count_e,len(self.data)), dtype=np.int64)
            self.av = np.zeros((self.count_c,self.count_e,len(self.data)), dtype=np.float64)
            for c in range(self.count_c):
                for e in range(self.count_e):
                    self.choice[c][e] = self.data["choice_" + str(c) + "_" + str(e)].values
                    self.av[c][e] = self.data["av_" + str(c) + "_" + str(e)].values


# -*- coding: utf-8 -*-
"""
Created on Fri Jan 29 16:39:30 2021

@author: j.reul

This module imports model parameters from pvrevious estimations to 
conduct simulations.
"""

import os
import pickle

PATH_MODULE = os.path.dirname(__file__)
sep = os.path.sep
PATH_ModelParam = PATH_MODULE + sep + 'ModelParam' + sep

# Load parameter-files from the estimation
with open(PATH_ModelParam + "initial_point_" + "car_ownership.pickle", "rb") as handle:
    initial_point_cars = pickle.load(handle)

with open(PATH_ModelParam + "initial_point_mode.pickle", "rb") as handle:
    initial_point_mode = pickle.load(handle)
    
with open(PATH_ModelParam + "asc_offset_hh_cars.pickle", 'rb') as handle:
    asc_offset_hh_cars = pickle.load(handle)  
    
# load auxiliary data to sample distances of activities and to derive the average speed
with open(PATH_ModelParam + "speed_parameters.pickle", "rb") as handle:
    log_param = pickle.load(handle)
    
dict_specific_travel_cost = {1: 0,
    2: 0,
    3: 0.2062,
    4: 0.1473,
    5: 0.0617,
    6: 0.0184,
    7: 0.4653,
    8: 0.0283,
    9: 0.0608,
    10: 0.2546
    }  # 2020: status quo
cc_cost_2020 = 0.33  # €/min carsharing-Kosten


# -*- coding: utf-8 -*-
"""
Created on Thu Jul 21 09:55:20 2022

@author: j.reul
"""

#import necessary modules
import time
import numpy as np

import mode_behave as mb

#%% SIMPLE MODEL - 3 ATTRIBUTES

#define model parameters
param_fixed = []
param_random = ['RELATIVER_KAUFPREIS', 'REICHWEITE_DURCH100', 'LADE_TANK_ZEIT']

param_temp = {'constant': 
                  {
                   'fixed':[],
                   'random':[]
                   },
              'variable':
                  {
                   'fixed': [],
                   'random':[]
                   }
              }

param_temp['variable']['fixed'] = param_fixed
param_temp['variable']['random'] = param_random    
    
#Initialize model
model = mb.Core(
    param=param_temp, 
    data_name='example_data', 
    alt=4, 
    equal_alt=1
    )

#%% ESTIMATION OF PARAMETERS.

#estimate MXL model
start = time.time()
#estimate mixed logit model
model.estimate_mixed_logit(
    min_iter=10, 
    max_iter=1000,
    tol=0.01,
    space_method = 'std_value',
    scale_space = 2,
    max_shares = 1000,
    bits_64=True,
    t_stats_out=False
    )
end = time.time()
delta = int(end-start)
print('Estimation of logit model took: ', str(delta), ' seconds.')

#%%Evaluation of MNL- and MXL-model
print("LL-Ratio of MNL-model:", model.loglike_MNL()[0])
print("LL-Ratio of MXL-model:", model.loglike_MXL())

#%%Post-processing
# Visualization of MXL-results and indication of clustering-results. 
# Play around with bw_adjust to adapt curve-smoothing.
model.visualize_space(
    k=2, 
    scale_individual=True, 
    cluster_method='kmeans', 
    external_points=np.array([model.initial_point]),
    bw_adjust=0.03,
    names_choice_options={0: "ICEV", 1: "PHEV", 2: "BEV", 3: "FCEV"}
    )
#%%
# Comment on forecasting: The forecasting is performed on training data 
# for simplicity. Please adjust the data (self.data) for further/own 
# forecasting practices by reloading a diverging subset of the base data.

# Forecast of cluster centers on the base data
model.forecast(method='LC', 
            k=2,
            cluster_method='kmeans',
            names_choice_options={0: "ICEV", 1: "PHEV", 2: "BEV", 3: "FCEV"},
            name_scenario='clustering'
            )
#%%
# Forecast of MNL model with parameter variation.
# In this example, the relative purchase price of vehicle technologies is 
# adjusted: ICEV +10%, PHEV +10%, BEV -50%, FCEV -50%
model.forecast(method='MNL', 
            sense_scenarios={"Cheap_EV": {
                "RELATIVER_KAUFPREIS": [[1.1], [1.1], [0.5], [0.5]]}
                },
            names_choice_options={0: "ICEV", 1: "PHEV", 2: "BEV", 3: "FCEV"},
            name_scenario='sensitivity'
            )


from setuptools import setup

setup(name='mode_behave',
      version='1.0.3',
      description='Estimation and simulation of discrete choice models',
      author='Julian Reul',
      author_email='j.reul@fz-juelich.de',
      url='https://github.com/julianreul/mode_behave',
      license='MIT',
      packages=['mode_behave_public'],
      include_package_data=True,
      zip_safe=False)


# -*- coding: utf-8 -*-
"""
Created on Mon Jan 27 11:20:18 2020

@author: j.reul
"""

from .core import Core
from .estimation import Estimation
from .simulation import Simulation
from .post_analysis import PostAnalysis


# -*- coding: utf-8 -*-
"""
Created on Sat May  7 11:46:16 2022

@author: j.reul
"""

import unittest
import numpy as np

import mode_behave as mb

class TestCarOwnership(unittest.TestCase):
    def test_simulate_hh_cars(self):
        """
        Integration test of the method -simulate_hh_cars()-

        """
        #Define the input data
        urban_region = 1
        rural_region = 0
        hh_size = 2
        adults_working = 1 
        children = 0
        htype = 1
        quali_opnv = 2 
        sharing = 1
        relative_cost_per_car = 2 
        age_adults = 3.5
        regiontype = 1
        
        #simulate
        model = mb.Core(model_type = 'simulation')        
        result = model.simulate_hh_cars(urban_region, rural_region, hh_size,
                             adults_working, children, htype, quali_opnv, sharing,
                             relative_cost_per_car, age_adults, regiontype)
                 
        #test
        self.assertTrue(
            np.allclose(
            result, 
            [0.500314856405928,
             0.43400343757783405,
             0.06463228081919556,
             0.0010494251970423272],
            atol=0.01)
            )      
        
    def test_simulate_mode_choice(self): 
        """
        Integration test of the method -test_simulate_mode_choice()-.

        """
        #simulate
        model = mb.Core(model_type = 'simulation')        
        result = model.simulate_mode_choice(
            agegroup=2,
            occupation=1,
            regiontype=1,
            distance=10,
            av=np.array([1]*10)
            )
                 
        #test
        self.assertTrue(
            np.allclose(
            result,
            [0.004245231161460397,
             0.03591730407748586,
             0.37137115355105155,
             0.34324035236575373,
             0.04877611038816796,
             0.04172907422256176,
             0.1287949180524477,
             0.017423477365014716,
             0.008488917395244943,
             1.3461420811290233e-05],
            atol=0.01)
            )    
                         
if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
"""
Created on Sat May  7 11:46:35 2022

@author: j.reul
"""

import unittest
import pickle
import numpy as np
import pandas as pd
from scipy import stats

import mode_behave as mb

class TestEstimation(unittest.TestCase):       
        
    def get_artificial_data(self):
        """
        This method was used to create the artificial dataset, upon which
        the tested estimation routines below are running.

        Raises
        ------
        AttributeError
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
                
        size_dataset = 1000
        
        #DEFINE COPULA
        #mean values
        mu_c = np.array([0,0,0,0])
        #Covariance between attributes and choice
        psi_choice = 0.5
        psi_between = 0.8
        dist_attr = "weibull"
        #risks: Choice, attr_1, attr_2, attr_3
        cov_c = np.array(
            [
                [1, psi_choice, psi_choice, psi_choice],
                [psi_choice, 1, psi_between, psi_between],
                [psi_choice, psi_between, 1, psi_between],
                [psi_choice, psi_between, psi_between, 1],
                ]
            )
        dist_c = stats.multivariate_normal(mean=mu_c, cov=cov_c)
        #____obtain random sample from copula distribution
        sample_c = dist_c.rvs(size=size_dataset)
        #____obtain marginals from copula distribution
        choice = sample_c[:,0]
        attr_x = sample_c[:,1]
        attr_y = sample_c[:,2]
        attr_z = sample_c[:,3]
        
        #TRANSFORM TO UNIFORM
        u_choice = stats.norm.cdf(choice)
        u_attr_x = stats.norm.cdf(attr_x)
        u_attr_y = stats.norm.cdf(attr_y)
        u_attr_z = stats.norm.cdf(attr_z)
        
        #DEFINE MARGINAL DISTRIBUTIONS
        
        dist_choice = stats.uniform()
        if dist_attr == "uniform":
            dist_attr_x = stats.uniform()
            dist_attr_y = stats.uniform()
            dist_attr_z = stats.uniform()
        elif dist_attr == "weibull":
            dist_attr_x = stats.weibull_min(c=1.5)
            dist_attr_y = stats.weibull_min(c=1.1)
            dist_attr_z = stats.weibull_min(c=0.5)
        else:
            raise AttributeError("unknown dist_attr.")
            
        #DRAW VECTOR FROM COPULA
        beta = np.array(list(zip(
            dist_choice.ppf(u_choice), 
            dist_attr_x.ppf(u_attr_x), 
            dist_attr_y.ppf(u_attr_y),
            dist_attr_z.ppf(u_attr_z), 
            )))
        
        #Create pandas dataframe out of beta
        dataset = pd.DataFrame(data=beta, columns=["choice_float", "attr_x", "attr_y", "attr_z"])
        
        #def get choice int
        def get_choice_int(choice_float):
            if choice_float < 0.33:
                return 0
            elif choice_float < 0.66:
                return 1
            else:
                return 2
            
        dataset["choice_int"] = dataset["choice_float"].apply(get_choice_int)
        dataset["choice_0"] = (dataset["choice_int"] == 0)*1
        dataset["choice_1"] = (dataset["choice_int"] == 1)*1
        dataset["choice_2"] = (dataset["choice_int"] == 2)*1
        
        #create columns for attributes
        dataset["attr_x_0"] = dataset["attr_x"].copy()
        dataset["attr_x_1"] = dataset["attr_x"].copy()
        dataset["attr_x_2"] = dataset["attr_x"].copy()
        
        dataset["attr_y_0"] = dataset["attr_y"].copy()
        dataset["attr_y_1"] = dataset["attr_y"].copy()
        dataset["attr_y_2"] = dataset["attr_y"].copy()
        
        dataset["attr_z_0"] = dataset["attr_z"].copy()
        dataset["attr_z_1"] = dataset["attr_z"].copy()
        dataset["attr_z_2"] = dataset["attr_z"].copy()
        
        #create columns for availability
        dataset["av_0"] = 1
        dataset["av_1"] = 1
        dataset["av_2"] = 1
        
        #delete unnecessary columns
        dataset = dataset.drop(columns=[
            "choice_float",
            "attr_x", 
            "attr_y",
            "attr_z",
            "choice_int"
            ])
        
        #append "_0" to each column to fit required data format.
        for col in dataset.columns:
            dataset = dataset.rename(columns={col : col + "_0"})
            
        return dataset
            
    
    def test_estimate_mixed_logit(self):
        """
        Integration test of the method -estimate_mixed_logit()-
        Utilized CPU hardware only.

        """
        param_fixed = []
        param_random = [
            "attr_x",
            "attr_y",
            "attr_z"
            ]
        
        param_temp = {'constant': 
                          {
                           'fixed':[],
                           'random':[]
                           },
                      'variable':
                          {
                           'fixed': [],
                           'random':[]
                           }
                      }
        
        param_temp['variable']['fixed'] = param_fixed
        param_temp['variable']['random'] = param_random   
            
        #Initialize model
        model = mb.Core(
            param=param_temp, 
            data_name="artificial_data", 
            alt=3,
            equal_alt=1,
            include_weights=False,
            )
        
        #estimate mixed logit model
        model.estimate_mixed_logit(
            min_iter=10, 
            max_iter=1000,
            tol=0.01,
            space_method = 'std_value',
            scale_space = 2,
            max_shares = 1000,
            bits_64=True,
            t_stats_out=False
            )
        
        with open(model.PATH_ModelParam + "initial_point_artificial_data.pickle", 'rb') as handle:
            initial_point_compare = pickle.load(handle)  
                
        with open(model.PATH_ModelParam + "shares_artificial_data.pickle", 'rb') as handle:
            shares_compare = pickle.load(handle)  
            
        with open(model.PATH_ModelParam + "points_artificial_data.pickle", 'rb') as handle:
            points_compare = pickle.load(handle)  
        
        #test estimation of initial_point (via the method estimate_logit())
        self.assertTrue(np.allclose(model.initial_point, initial_point_compare, atol=0.1))
        
        #test estimation of shares
        self.assertTrue(np.allclose(model.shares, shares_compare, atol=0.1))
        
        #test definition of parameter space (points)
        self.assertTrue(np.allclose(model.points, points_compare, atol=0.1))

        
    def test_estimate_logit(self):
        """
        Integration test of the method -estimate_logit()-

        """
        
        param_fixed = []
        param_random = [
            "attr_x",
            "attr_y",
            "attr_z"
            ]
        
        param_temp = {'constant': 
                          {
                           'fixed':[],
                           'random':[]
                           },
                      'variable':
                          {
                           'fixed': [],
                           'random':[]
                           }
                      }
        
        param_temp['variable']['fixed'] = param_fixed
        param_temp['variable']['random'] = param_random         
            
        #Initialize model
        model = mb.Core(
            param=param_temp, 
            data_name="artificial_data", 
            alt=3,
            equal_alt=1,
            include_weights=False,
            )
                 
        model.initial_point = model.estimate_logit(stats=False)
        
        with open(model.PATH_ModelParam + "initial_point_artificial_data.pickle", 'rb') as handle:
            initial_point_compare = pickle.load(handle)  
                
        #test estimation of initial_point (via the method estimate_logit())
        self.assertTrue(np.allclose(model.initial_point, initial_point_compare, atol=0.1))
               
                         
if __name__ == '__main__':
    unittest.main()